/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals self, postMessage, importScripts, Map, Player, ROT, TECHS, ALLTECHS, ShipDesign, ShipDesignType */
/* globals ResearchOrder, CarryForwardOrder, ImproveColonyOrder, AddMarinesOrder, RegisterShipDesign, NullOrder */
/* globals BuyVehicleOrder, BuyStarBaseOrder */
/* globals ShipOrder, RepairShipOrder, ScrapShipOrder */
/* ------- console */
/* exported onmessage */

if (!self.document) {
	importScripts('map.js?' + location.search);
	importScripts('ship.js?' + location.search);
	importScripts('shipdesign.js?' + location.search);
	importScripts('shiporder.js?' + location.search);
	importScripts('tech.js?' + location.search);
	importScripts('player.js?' + location.search);
	importScripts('order.js?' + location.search);
	importScripts('../rot/rot.min.js?' + location.search);

	ShipDesign.init();

	const STATES = ["explore","expand","exploit","exterminate","invade","defend","research"];

	var onmessage = function(message) {
		const datastylename = message.data.stylename;
		const databasemap = JSON.parse(message.data.baseMap);
		const datacapital = message.data.capital;
		const datatech = message.data.tech;
		const dataplayer = JSON.parse(message.data.player);

		const turn = parseInt(message.data.turn);
		const basemap = new Map(databasemap);
		const player = new Player(datastylename,datacapital,datatech);

		player.fromJSON(dataplayer,basemap);
		if (player.seed) {
			ROT.RNG.setSeed(player.seed); // Set ROT's seed on this computer worker thread
			for(let i=1; i<10; i++) {
				ROT.RNG.getUniform(); // Pull some numbers to get into a more random space
			}
		} else {
			ROT.RNG.setState(JSON.parse(player.ROTRNGState));
		}

		let orders = [new NullOrder()];
		let shiporders = [];

		if (player.active) {
			player.computeCommsGrid();
			player.state = newState(turn,player);

			moveShips(player);
			orders = spendMoney(player);

			for(let ship of player.ships) {
				shiporders.push(ship.order);
			}
		}

		let reply = {orders: JSON.stringify(orders),
					 ships: JSON.stringify(shiporders),
					 state: JSON.stringify(player.state),
					 designid: player.designid,
					 ROTRNGState: JSON.stringify(ROT.RNG.getState()) };
		postMessage(reply);
	};

	var newState = function(turn, player) {
		const oldState = player.state;
		if (oldState && ROT.RNG.getPercentage() > 15) return oldState; // No need to change state

		let possibles = {};
		for(let actionName of STATES) {
			possibles[actionName] = 20;
		}
		// Scan the player's map - unscanned hexes add twenty to explore - add one per missed turn for other hexes to explore.
		// Uncolonised worlds add 50 to expand
		player.map.forEachHex(hex => {
			if (hex.onGrid) {
				if (hex.lastScanned) {
					if (hex.planet && hex.level == 0) {
						possibles.expand += 50;
					}
					possibles.explore += (hex.lastScanned - turn);
				} else {
					possibles.explore += 20;
				}
			}
		});
		const newState = ROT.RNG.getWeightedValue(possibles);
		/*if (newState != oldState) {
			console.info(`${player.stylename} changing to ${newState}`);
		}*/
		return newState;
	};

	var moveShips = function(player) {
		var shipyards = [];
		player.map.forEachHex(hex => {
			if (hex.owner == player.stylename && hex.planet && hex.level && hex.level > 19) {
				shipyards.push(hex.id);
			}
		});
		for(let ship of player.ships) {
			const selfShipyards = shipyards;
			const selfShip = ship;
			if (ship.retreatingTo) {
				ship.order = new ShipOrder(ship.retreatingTo);
			}
			else if (ship.getRepairCost() > 0 && selfShipyards.some(hexid => hexid == selfShip.location) && ROT.RNG.getPercentage() < 51) {
				ship.order = new RepairShipOrder(ship.getRepairCost());
			} else if (ROT.RNG.getPercentage() < 10) {
				ship.order = new ScrapShipOrder(ship.getScrapValue());
			} else {
				ship.order = new ShipOrder();
				if	(ship.getJumpRange() > 0 &&
					(	(ROT.RNG.getPercentage() > 40) ||
						(player.state == "explore" && ROT.RNG.getPercentage() > 25))) {
					const possibleHexes = [];
					for(let row=1; row<21; row++) {
						for(let column of "ABCDEFGHIJKLMNOP") {
							const name = column+row;
							if (Map.range(ship.location,name) <= ship.getJumpRange() && player.map.hex(name).onGrid) {
								possibleHexes.push(name);
							}
						}
					}
					ship.order.jump = ROT.RNG.getItem(possibleHexes);
				}
				if (ship.getColonyImprovement() > 0 && (ROT.RNG.getPercentage() > 40)) {
					ship.order.colonise = true;
				}
				if (ship.marines > 0 && ROT.RNG.getPercentage() > 50) {
					ship.order.setMarinesToDisembark();
				}
				if (ship.marines < ship.getMaxMarines() && ROT.RNG.getPercentage() > 50) {
					ship.order.setMarinesToEmbark();
				}
			}
		}
	};

	var evaluateDesign = function(design,player) {
		if (design === null) return 0;
		switch(player.state) {
			case "explore": return (design.getJumpRange() > player.researchLevel(TECHS.communications) ? player.researchLevel(TECHS.communications) : design.getJumpRange());
			case "expand":  return design.getRatingColony();
			case "exterminate": return design.getRatingAttack();
			case "invade": return design.getRatingTroops();
			case "defend": return design.getRatingDefence();
			default: return 1;
		}
	};

	var designNewShip = function(player,type,budget) {
		if (budget < 100) return null; // The cheapest possible ship is 50 PPs (which will be doubled for the first one) so if the proposed budget isn't even that high - give up now!
		if (player.state == "exploit") return null; // Not going to bother if we are supposed to be exploiting our resources
		if (player.state == "research") return null; // Not going to bother if we are supposed to be researching
		if (player.state == "explore" && player.researchLevel(TECHS.communications) <3) return null; // The TL1 scout is already Jump-2
		if (player.state == "expand" && player.researchLevel(TECHS.communications) < 2 && player.researchLevel(TECHS.shipSize) == 1) return null; // The TL1 colony ship is already Jump-1
		if (player.state == "invade" && player.researchLevel(TECHS.communications) < 2  && player.researchLevel(TECHS.shipSize) == 1) return null; // The TL1 troop transport is already Jump-1
		// First what's the best ship design so far
		let currentDesignsInBudget = Object.values(ShipDesign.REGISTRY)
			.filter(design => design.type == type)
			.filter(design => design.isTechSufficent({ power:	player.researchLevel(TECHS.powerPlant),
					jump:		player.researchLevel(TECHS.jumpDrive),
					man:		player.researchLevel(TECHS.maneuverDrive),
					size:		player.researchLevel(TECHS.shipSize),
					weapons:	player.researchLevel(TECHS.weaponary),
					defences:	player.researchLevel(TECHS.shipDefences),
					computer:	player.researchLevel(TECHS.computers),
					arch:		player.researchLevel(TECHS.archaeology),
					planet:	player.researchLevel(TECHS.planetaryDefences)}))
			.filter(design => design.totalCost()*builtFactor(player,design,[]) <= budget)
			.sort((a,b) => {
				return evaluateDesign(b,player) - evaluateDesign(a,player);
			});
		const currentBest = (currentDesignsInBudget.length ? currentDesignsInBudget[0] : null);
		const currentBestScore = evaluateDesign(currentBest,player);
		//console.info(`designNewShip: ${player.stylename} budget=${budget} priority=${player.state} currentBest=${currentBest ? currentBest.name : "none"} score=${currentBestScore}`);
		let bestProposedDesign = null;
		let bestScore = 0;
		for(let i=0; i<10; i++) {
			const proposedDesign = getNewRandomVesselDesign(player,type,budget);
			if (proposedDesign == null) continue; // The design failed (probably too large!)
			if (proposedDesign.totalCost()*builtFactor(player,proposedDesign,[]) > budget) {
				proposedDesign.delete();
				continue; // Try another design
			}
			const proposedScore = evaluateDesign(proposedDesign,player);
			//console.info(`${player.stylename}: Possible New Design: ${JSON.stringify(proposedDesign)} scores ${proposedScore}`);
			if (proposedScore <= currentBestScore){
				proposedDesign.delete();
				continue; // Not good enough to beat the best design so far
			}
			if (proposedScore <= bestScore) {
				proposedDesign.delete(); // Remove from the registry as we're not going to use it
				continue; // Try another design
			}
			if (bestProposedDesign) {
				bestProposedDesign.delete();
			}
			bestProposedDesign = proposedDesign;
			bestScore = proposedScore;
		}
		if (bestProposedDesign) {
			//console.info(`${player.stylename}: New Design: ${JSON.stringify(bestProposedDesign)}`);
			return bestProposedDesign;
		}
		return null;
	};

	var getNewRandomVesselDesign = function(player,type,budget) {
		// Always start with a bridge
		let blocks = ["_b0","_b1","_b2","_b3","_b4","_b5","_b6","_b7","_b8"].slice(0,player.researchLevel(TECHS.computers)+1);
		const name = player.getNextDesignId();
		const design = new ShipDesign(name,type,name);
		design.addBlock(ROT.RNG.getItem(blocks));

		// Also always need a power for a useful vessel - so we will give it one now
		let powerBlocks = ["_p1","_p2","_p3","_p4","_p5","_p6","_p7"].slice(0,player.researchLevel(TECHS.powerPlant));
		design.addBlock(ROT.RNG.getItem(powerBlocks));

		// Now add in all the other block types
		blocks = blocks.concat(powerBlocks);
		if (type == ShipDesignType.Ship) {
			blocks = blocks.concat(["_ja","_jb","_jc","_jd","_je","_jf","_jg","_jh"].slice(0,player.researchLevel(TECHS.jumpDrive)));
			blocks = blocks.concat(["_ma","_mb","_mc","_md","_me","_mf","_mg","_mh"].slice(0,player.researchLevel(TECHS.maneuverDrive)));

			if (player.state == "expand") {
				blocks.push("_cb");
				design.addBlock("_cb"); // It'll need at least one to stand a chance of being a 'good' design
			}
			if (player.state == "invade") {
				blocks.push("_bb");
				design.addBlock("_bb"); // It'll need at least one (and probably more!) to stand a chance of being a 'good' design
			}
		}
		blocks = blocks.concat(["_wl","_wp","_wa","_wf","_wr","_wc","_ws","_wm","_wt","_wx","_wb"].slice(0,player.researchLevel(TECHS.weaponary)));
		blocks = blocks.concat(["_da","_dc","_ds","_df","_di","_dg","_dr","_di"].slice(0,player.researchLevel(TECHS.shipDefences)));
		if (player.researchLevel(TECHS.archaeology) > 1) blocks = blocks.concat(["_an"]);
		if (player.researchLevel(TECHS.archaeology) > 2) blocks = blocks.concat(["_ad"]);
		if (player.researchLevel(TECHS.archaeology) > 3) blocks = blocks.concat(["_at"]);
		if (player.researchLevel(TECHS.archaeology) > 4) blocks = blocks.concat(["_as"]);

		const maxSizes = [0,10,20,30,40,60,80,100,150,1e9];
		const maxSize = maxSizes[player.researchLevel(TECHS.shipSize)];
		while(design.totalSize() < maxSize && (ROT.RNG.getUniform() > (2/maxSize)) && design.totalCost()*builtFactor(player,design,[]) < budget) {
			design.addBlock(ROT.RNG.getItem(blocks));
		}
		if (design.totalSize() <= maxSize && design.totalCost()*builtFactor(player,design,[]) < budget) {
			return design;
		} else {
			design.delete();
			return null; // Design is not valid
		}
	};

	var spendMoney = function(player) {
		const orders = [];
		const colonies = [];
		const allHexesOnGrid = [];
		let money = player.production+player.carriedProduction;
		player.map.forEachHex(function(hex) {
			if (hex.owner == player.stylename && hex.level > 0) {
				colonies.push(hex.id);
			}
			if (hex.onGrid) {
				allHexesOnGrid.push(hex.id);
			}
		});

		if (ROT.RNG.getPercentage() < 21) {
			const design = designNewShip(player,ShipDesignType.Ship,money);
			if (design) {
				const order =  new RegisterShipDesign(design);
				const colony = ROT.RNG.getItem(colonies.filter(colony => player.map.hex(colony).level > 19));
				const name = design.name+" "+player.stylename+"-"+getRandom(0,99999999);
				const buyorder = new BuyVehicleOrder(design,name,colony);
				const amount = design.totalCost()*builtFactor(player,design,orders);
				orders.push(order);
				orders.push(buyorder);
				money -= amount;
			}
		}

		if (ROT.RNG.getPercentage() < 11 && player.researchLevel(TECHS.planetaryDefences) == 3) {
			const design = designNewShip(player,ShipDesignType.SpaceBase,money);
			if (design) {
				const order =  new RegisterShipDesign(design);
				const colony = ROT.RNG.getItem(colonies.filter(colony => player.map.hex(colony).level > 19));
				const name = design.name+" "+player.stylename+"-"+getRandom(0,99999999);
				const buyorder = new BuyVehicleOrder(design,name,colony);
				const amount = design.totalCost()*builtFactor(player,design,orders);
				orders.push(order);
				orders.push(buyorder);
				money -= amount;
			}
		}

		if (ROT.RNG.getPercentage() < 11 && player.researchLevel(TECHS.planetaryDefences) > 3) {
			const design = designNewShip(player,ShipDesignType.StarBase,money);
			if (design) {
				const order =  new RegisterShipDesign(design);
				const colony = ROT.RNG.getItem(colonies.filter(colony => player.map.hex(colony).level > 19));
				const name = design.name+" "+player.stylename+"-"+getRandom(0,99999999);
				const buyorder = new BuyVehicleOrder(design,name,colony);
				const amount = design.totalCost()*builtFactor(player,design,orders);
				orders.push(order);
				orders.push(buyorder);
				money -= amount;
			}
		}

		let spendMoneyOnResearch = function(_money,_priority) {
			const amount = getRandom(1,money);
			const tech = ROT.RNG.getItem(ALLTECHS);
			const order = new ResearchOrder(tech,tech.index === player.techIndex ? amount*2 : amount);
			return [order,amount];
		};
		let improvedColonies = {};
		let spendMoneyOnImprovingColony = function(_money,_priority) {
			const colony = ROT.RNG.getItem(colonies);
			let order = null;
			let amount = 0;
			if (player.map.hex(colony).level > 4 && player.map.hex(colony).level <= money/20) {
				if (improvedColonies[colony]) {
					// No order
				} else {
					order = new ImproveColonyOrder(colony);
					amount = player.map.hex(colony).level*20;
					improvedColonies[colony] = true;
				}
			}
			return [order,amount];
		};
		let marinedColonies = {};
		let spendMoneyOnMarines = function(money,_priority) {
			let order = null;
			let amount = 0;
			if (money > 9) {
				const colony = ROT.RNG.getItem(colonies);
				if (marinedColonies[colony]) {
					// No order
				} else {
					const companies = getRandom(1,Math.floor(money/10));
					order = new AddMarinesOrder(colony,companies);
					amount = companies*10;
					marinedColonies[colony] = true;
				}
			}
			return [order,amount];
		};
		let spendMoneyOnShip = function(money,_priority) {
			let order = null;
			let amount = 0;
			const selfPlayer = player;
			const selfMoney = money;
			const selfBuiltFactor = builtFactor;
			const colony = ROT.RNG.getItem(colonies.filter(colony => selfPlayer.map.hex(colony).level > 19));
			const possibleDesigns = Object.values(ShipDesign.REGISTRY)
				.filter(design => design.type !== ShipDesignType.StarBase)
				.filter(design => design.isTechSufficent({ power:	selfPlayer.researchLevel(TECHS.powerPlant),
					jump:		selfPlayer.researchLevel(TECHS.jumpDrive),
					man:		selfPlayer.researchLevel(TECHS.maneuverDrive),
					size:		selfPlayer.researchLevel(TECHS.shipSize),
					weapons:	selfPlayer.researchLevel(TECHS.weaponary),
					defences:	selfPlayer.researchLevel(TECHS.shipDefences),
					computer:	selfPlayer.researchLevel(TECHS.computers),
					arch:		selfPlayer.researchLevel(TECHS.archaeology),
					planet:	selfPlayer.researchLevel(TECHS.planetaryDefences)}))
				.filter(design => design.totalCost()*selfBuiltFactor(selfPlayer,design,orders) <= selfMoney);
			if (possibleDesigns.length > 0) {
				const design = ROT.RNG.getItem(possibleDesigns);
				const name = design.name+" "+player.stylename+"-"+getRandom(0,99999999);
				order = new BuyVehicleOrder(design,name,colony);
				amount = design.totalCost()*builtFactor(player,design,orders);
			}
			return [order,amount];
		};
		let spendMoneyOnStarBase = function(money,_priority) {
			let order = null;
			let amount = 0;
			const location = ROT.RNG.getItem(allHexesOnGrid);
			const selfPlayer = player;
			const selfMoney = money;
			const selfBuiltFactor = builtFactor;
			const possibleDesigns = Object.values(ShipDesign.REGISTRY)
				.filter(design => design.type === ShipDesignType.StarBase)
				.filter(design => design.isTechSufficent({ power:	selfPlayer.researchLevel(TECHS.powerPlant),
					jump:		selfPlayer.researchLevel(TECHS.jumpDrive),
					man:		selfPlayer.researchLevel(TECHS.maneuverDrive),
					size:		selfPlayer.researchLevel(TECHS.shipSize),
					weapons:	selfPlayer.researchLevel(TECHS.weaponary),
					defences:	selfPlayer.researchLevel(TECHS.shipDefences),
					computer:	selfPlayer.researchLevel(TECHS.computers),
					arch:		selfPlayer.researchLevel(TECHS.archaeology),
					planet:	selfPlayer.researchLevel(TECHS.planetaryDefences)}))
				.filter(design => design.totalCost()*selfBuiltFactor(selfPlayer,design,orders) <= selfMoney);
			if (possibleDesigns.length > 0) {
				const design = ROT.RNG.getItem(possibleDesigns);
				const name = design.name+" "+player.stylename+"-"+getRandom(0,99999999);
				order = new BuyStarBaseOrder(design,name,location);
				amount = design.totalCost()*builtFactor(player,design,orders);
			}
			return [order,amount];
		};

		// Decide on where to spend money
		let functions = {research: spendMoneyOnResearch,
			improveColony: spendMoneyOnImprovingColony,
			marines: spendMoneyOnMarines,
			buildShip: spendMoneyOnShip,
			buildStarBase: spendMoneyOnStarBase
		};
		let options = {research: 34, improveColony: 16, marines: 16, buildShip: 39, buildStarBase: 5 };
		if (player.state == "explore") options.buildShip += 20;
		if (player.state == "research") options.research += 20;
		if (player.state == "exploit") options.improveColony += 20;
		if (player.state == "defend") { options.research += 10; options.marines += 10; options.buildStarBase += 10; }
		if (player.state == "exterminate") options.buildShip += 20;
		if (player.state == "invade") { options.marines += 10; options.buildShip += 10; }

		while(money > 0 && ROT.RNG.getPercentage() > 16) {
			const choice = ROT.RNG.getWeightedValue(options);
			const [newOrder,deduction] = [ ...functions[choice](money,player.state) ];
			if (newOrder) {
				orders.push(newOrder);
				money -= deduction;
			}
		}
		orders.push(new CarryForwardOrder(money));
		//console.info(`${player.stylename} orders are [${orders}]`);
		return orders;
	};

	var builtFactor = function(player,design,orders) {
		let number = design.numberOfClassBuilt(player)+1;
		// Scan the orders looking for orders of this ship class - if so bump the number by one
		for(let order of orders) {
			if ((order instanceof BuyVehicleOrder) && (order.design === design)) { number++; }
		}
		// If the number is 1 AND this is a second or later ship we must double the price
		if (number === 1 && hasEverBuiltAShip(player,orders)) {
			return 2;
		}
		return 1;
	};

	var hasEverBuiltAShip = function(player,orders) {
		if (player.hasBuiltAShip) return true;
		for(let order of orders) {
			if (order instanceof BuyVehicleOrder) return true;
		}
		return false;

	};

	var getRandom = function(min,max) {
		return Math.floor(ROT.RNG.getUniform()*(max-min+1))+min;
	};
}